/*******************************************************************************
 *	@file			Tetris.cpp
 *	@brief			egX
 *
 *	@date			10-Feb-2011(Tue)
 ******************************************************************************/

// vvZbT---------------------------------------------------------------
#include "DL.h"
#include "DLGraph.h"
#include "DLSound.h"
#include "DLInput.h"
#include "DLDebug.h"

#ifdef _DEBUG
#pragma comment ( lib, "Debug\\_DXLibv2.lib" )
#else	// _DEBUG
#pragma comment ( lib, "Release\\_DXLibv2.lib" )
#endif	// _DEBUG

// 񋓌^`-------------------------------------------------------------------

// \̒`-------------------------------------------------------------------

// O[oϐ-----------------------------------------------------------
static const char* const APPNAME	= "TETRIS";									//!< Av̖O
static const int APPWIDTH			= 640;										//!< Av̏
static const int APPHEIGHT			= 480;										//!< Av̏
static const float FRAMERATE		= 33.33333333333f;							//!< t[[g

DL::CDLGraph*			g_pGraph;												//!< DirectX Graphics
DL::CDLSound*			g_pSound;												//!< DirectX Sound
DL::CDLInput*			g_pInput;												//!< DirectX Input
DL::CDLDebug*			g_pDebug;

HANDLE					g_hMutex = NULL;										//!< Lnh
HWND					g_hWnd = NULL;											//!< EChEnh

BYTE					g_byKey[ DL::CDLInput::KEYBUFFER_MAX ];					//!< L[obt@

// -----------------------------------------------------------------------------
// egXf[^
static const int		TETRIS_WIDTH	= 12;									//!< PQubN
static const int		TETRIS_HEIGHT	= 21;									//!< cQPubN
static const int		TETRIS_BLOCK	= 4;									//!< ubN̏c
static const int		TETRIS_LIST		= 7;									//!< ubNXg
static const int		TETRIS_SIZE		= 20;									//!< ubÑTCY
static const int		START_BLOCKX	= 4;									//!< ubNJnʒu
static const int		START_BLOCKY	= 0;									//!< ubNJnʒu

// ubŇŒf[^
static const int		BLOCK_LIST[ TETRIS_LIST ][ TETRIS_BLOCK ][ TETRIS_BLOCK ] =
{
	{	0, 1, 0, 0,		0, 1, 0, 0,		0, 1, 0, 0,		0, 1, 0, 0,	},
	{	0, 0, 0, 0,		0, 2, 2, 0,		0, 2, 0, 0,		0, 2, 0, 0,	},
	{	0, 0, 3, 0,		0, 3, 3, 0,		0, 3, 0, 0,		0, 0, 0, 0,	},
	{	0, 4, 0, 0,		0, 4, 4, 0,		0, 0, 4, 0,		0, 0, 0, 0,	},
	{	0, 0, 0, 0,		0, 5, 0, 0,		5, 5, 5, 0,		0, 0, 0, 0,	},
	{	0, 0, 0, 0,		0, 6, 6, 0,		0, 6, 6, 0,		0, 0, 0, 0,	},
	{	0, 0, 0, 0,		0, 7, 7, 0,		0, 0, 7, 0,		0, 0, 7, 0,	},
};

// ubN̐F
static const int		BLOCK_COLOR[ 10 ] =
{
	0xff000000,	0xff888888,	0xff0000ff,
	0xff00ff00,	0xffff0000,	0xff00ffff,
	0xffffff00,	0xffff00ff,	0xff000000,
	0xffffffff,
};

int						g_nStage[ TETRIS_HEIGHT ][ TETRIS_WIDTH ];				//!< ǂƌŒς݃ubN
int						g_nBlock[ TETRIS_BLOCK ][ TETRIS_BLOCK ];				//!< 쒆̃ubN

int						g_nGameOver;											//!< Q[I[o[tO
int						g_nGameCnt;												//!< Q[JEg

int						g_nBlockType;											//!< ubN̎
int						g_nBlockX;												//!< ubN̂wʒu
int						g_nBlockY;												//!< ubN̂xʒu
int						g_nBaseX;												//!< wW
int						g_nBaseY;												//!< xW
BOOL					g_bKey[ 4 ];

int						g_nFontNum;


// ֐vg^Cv錾-----------------------------------------------------
void CreateBlock( void );														//!< ubN쐬
BOOL CheckOverlap( const int& nX, const int& nY );								//!< d
void MoveBlock( const int& nX, const int& nY );									//!< ubNړ
void TurnBlock( void );															//!< ubN]
void DropBlock( void );															//!< ubN
void LockBlock( void );															//!< ubNŒ
void CheckLine( void );															//!< 
void DrawTetris( void );														//!< `

// -----------------------------------------------------------------------------
/*******************************************************************************
 *	@fn				void CreateBlock( void )
 *	@retval			none
 *	@brief			ubN쐬
 ******************************************************************************/
void CreateBlock( void )
{
	// gpĎgpubN쐬
	::srand( static_cast<unsigned int>( ::timeGetTime() ) + g_nBlockType );
	g_nBlockType = ::rand() % TETRIS_LIST;

	// gpubNo^
	for( int i = 0; i < TETRIS_BLOCK; i++ )
		for( int j = 0; j < TETRIS_BLOCK; j++ )
		{
			g_nBlock[ i ][ j ] = BLOCK_LIST[ g_nBlockType ][ i ][ j ];
		}
}

/*******************************************************************************
 *	@fn				BOOL CheckOverlap( const int& nX, const int& nY )
 *	@retval			BOOL			TRUE=OK/FALSE=NG
 *	@brief			d
 ******************************************************************************/
BOOL CheckOverlap( const int& nX, const int& nY )
{
	for( int i = 0; i < TETRIS_BLOCK; i++ )
		for( int j = 0; j < TETRIS_BLOCK; j++ )
		{
			if( g_nBlock[ i ][ j ] != 0 )
			{
				if( g_nStage[ nY + i ][ nX + j ] != 0 )
					return FALSE;
			}
		}
	return TRUE;
}

/*******************************************************************************
 *	@fn				void MoveBlock( const int& nX, const int& nY )
 *	@retval			none
 *	@brief			ubNړ
 ******************************************************************************/
void MoveBlock( const int& nX, const int& nY )
{
	g_nBlockX = nX;
	g_nBlockY = nY;
}

/*******************************************************************************
 *	@fn				void TurnBlock( void )
 *	@retval			none
 *	@brief			ubN]
 ******************************************************************************/
void TurnBlock( void )
{
	// ꎞۑz
	int nTmp[ TETRIS_BLOCK ][ TETRIS_BLOCK ];

	// ꎞۑzɕۑ
	for( int i = 0; i < TETRIS_BLOCK; i++ )
		for( int j = 0; j < TETRIS_BLOCK; j++ )
		{
			nTmp[ i ][ j ] = g_nBlock[ i ][ j ];
		}

	// ubN]
	for( int i = 0; i < TETRIS_BLOCK; i++ )
		for( int j = 0; j < TETRIS_BLOCK; j++ )
		{
			g_nBlock[ i ][ j ] = nTmp[ 3 - j ][ i ];
		}

	// d`FbN
	if( !CheckOverlap( g_nBlockX, g_nBlockY ) )
	{
		// ]Ȃߌɖ߂
		for( int i = 0; i < TETRIS_BLOCK; i++ )
			for( int j = 0; j < TETRIS_BLOCK; j++ )
			{
				g_nBlock[ i ][ j ] = nTmp[ i ][ j ];
			}
	}
}

/*******************************************************************************
 *	@fn				void DropBlock( void )
 *	@retval			none
 *	@brief			ubN
 ******************************************************************************/
void DropBlock( void )
{
	// ړł
	if( CheckOverlap( g_nBlockX, g_nBlockY + 1 ) )
	{
		MoveBlock( g_nBlockX, g_nBlockY + 1 );
	}else
	// ړłȂ
	{
		// ubNŒ
		LockBlock();
		// 
		CheckLine();
		// VKubN쐬
		CreateBlock();
		g_nBlockX = START_BLOCKX;
		g_nBlockY = START_BLOCKY;
		// Q[I[o[
		if( !CheckOverlap( g_nBlockX, g_nBlockY ) )
			g_nGameOver = 1;
	}
}

/*******************************************************************************
 *	@fn				void LockBlock( void )
 *	@retval			none
 *	@brief			ubNŒ
 ******************************************************************************/
void LockBlock( void )
{
	for( int i = 0; i < TETRIS_BLOCK; i++ )
		for( int j = 0; j < TETRIS_BLOCK; j++ )
		{
			if( g_nBlock[ i ][ j ] != 0 )
			{
				g_nStage[ g_nBlockY + i ][ g_nBlockX + j ] = g_nBlock[ i ][ j ];
			}
		}
}

/*******************************************************************************
 *	@fn				void CheckLine( void )
 *	@retval			none
 *	@brief			
 ******************************************************************************/
void CheckLine( void )
{
	int nComp = 0;		// ̌
	int nLines = 0;		// ɏ

	for( int i = TETRIS_HEIGHT - 2; i > 0; i-- )
	{
		nComp = 1;
		for( int j = 1; j < TETRIS_WIDTH - 1; j++ )
		{
			// 󔒂
			if( g_nStage[ i ][ j ] == 0 )
			{
				nComp = 0;
				break;
			}
		}

		// 
		if( nComp )
		{
			nLines++;
			for( int j = 1; j < TETRIS_WIDTH - 1; j++ )
				g_nStage[ i ][ j ] = 0;

			// 炷
			for( int k = i; k > 0; k-- )
				for( int j = 1; j < TETRIS_WIDTH - 1; j++ )
					g_nStage[ k ][ j ] = g_nStage[ k - 1 ][ j ];

			// x
			i++;
		}
	}
}

/*******************************************************************************
 *	@fn				void DrawTetris( void )
 *	@retval			none
 *	@brief			`
 ******************************************************************************/
void DrawTetris( void )
{
	// ǂƌŒς݃ubN
	for( int i = 0; i < TETRIS_HEIGHT; i++ )
		for( int j = 0; j < TETRIS_WIDTH; j++ )
		{
			int n = g_nStage[ i ][ j ];

			DL::D3DOBJECT* m_pD3DObj = g_pGraph->GetObject( i * TETRIS_WIDTH + j );
			DL::D3DDATA* pData = static_cast<DL::D3DDATA*>( m_pD3DObj->m_pObj );

			DL::D3DVERTEX* v0 = static_cast<DL::D3DVERTEX*>( &pData->m_pData[ 0 ] );

			if( g_nGameOver && n != 0 ) n = 9;
			v0->color = BLOCK_COLOR[ n ];
		}

	// 쒆ubN
	for( int i = 0; i < TETRIS_BLOCK; i++ )
		for( int j = 0; j < TETRIS_BLOCK; j++ )
		{
			int n = g_nBlock[ i ][ j ];

			// `悵Ȃ
			if( n == 0 ) continue;

			DL::D3DOBJECT* m_pD3DObj = g_pGraph->GetObject( ( g_nBlockY + i ) * TETRIS_WIDTH + ( g_nBlockX + j ) );
			DL::D3DDATA* pData = static_cast<DL::D3DDATA*>( m_pD3DObj->m_pObj );

			DL::D3DVERTEX* v0 = static_cast<DL::D3DVERTEX*>( &pData->m_pData[ 0 ] );

			if( g_nGameOver ) n = 9;
			v0->color = BLOCK_COLOR[ n ];
		}
}

/*******************************************************************************
 *	@fn				LRESULT CALLBACK WndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
 *	@retval			LRESULT
 *	@brief			EChEvV[W
 ******************************************************************************/
LRESULT CALLBACK WndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	switch( msg )
	{
	case WM_DESTROY:															//!< EChE
		::PostQuitMessage( 0 );													//!< WM_QUIT𑗂
		return ::DefWindowProc( hWnd, msg, wParam, lParam );
	default :
		return ::DefWindowProc( hWnd, msg, wParam, lParam );
	}
	return ::DefWindowProc( hWnd, msg, wParam, lParam );
}

/*******************************************************************************
 *	@fn				int Init( const HINSTANCE& hInst )
 *	@param[in]		hInst			CX^Xnh
 *	@retval			int				EXIT_SUCCESS=OK/EXIT_FAILURE=NG
 *	@brief			AvP[V̏
 ******************************************************************************/
int Init( const HINSTANCE& hInst )
{
	// dŇ
	// L̊m
	g_hMutex = ::CreateMutex( NULL, FALSE, APPNAME );
	// Ľ
	if( ::WaitForSingleObject( g_hMutex, 0 ) != WAIT_OBJECT_0 )
	{
		_ASSERT( 0 && "WaitForSingleObject()" );
		::CloseHandle( g_hMutex );
		return EXIT_FAILURE;
	}

	// EChE
	WNDCLASS			wc;														//!< EChENX
	// EChENX̓o^
	wc.style				= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc			= static_cast<WNDPROC>( WndProc );					//!< bZ[W󂯎EChE֐
	wc.cbClsExtra			= 0;
	wc.cbWndExtra			= 0;
	wc.hInstance			= hInst;
	wc.hIcon				= NULL;												//!< ACRo^
	wc.hCursor				= ::LoadCursor( NULL, IDC_ARROW );					//!< J[\o^
	wc.hbrBackground		= static_cast<HBRUSH>( ::GetStockObject( WHITE_BRUSH ) );
	wc.lpszMenuName			= NULL;												//!< j[̓o^
	wc.lpszClassName		= APPNAME;
	// EChENX̓o^
	if( !::RegisterClass( &wc ) ) return EXIT_FAILURE;

	int nSizeX = APPWIDTH;
	int nSizeY = APPHEIGHT;
	int nPosX = ( ::GetSystemMetrics( SM_CXSCREEN ) - nSizeX ) / 2;;
	int nPosY = ( ::GetSystemMetrics( SM_CYSCREEN ) - nSizeY ) / 2;;
	// EChE쐬
	g_hWnd = ::CreateWindow( APPNAME,											//!< NX
							 APPNAME,											//!< ^Cg
							 WS_POPUP,											//!< X^C( WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU )
							 nPosX,												//!< wW
							 nPosY,												//!< xW
							 nSizeX,											//!< 
							 nSizeY,											//!< 
							 NULL,												//!< eEChẼnh
							 NULL,												//!< j[nh
							 hInst,												//!< CX^Xnh
							 NULL );											//!< lParam

	// EChE̍쐬
	if( !g_hWnd ) return EXIT_FAILURE;
	// EChE̕\Ԃ̐ݒ
	::ShowWindow( g_hWnd, SW_SHOW );
	// EChE̍XV
	::UpdateWindow( g_hWnd );

	return EXIT_SUCCESS;
}

/*******************************************************************************
 *	@fn				int Exit( void )
 *	@retval			int				EXIT_SUCCESS=OK/EXIT_FAILURE=NG
 *	@brief			AvP[V̏I
 ******************************************************************************/
int Exit( void )
{
	if( g_hWnd != NULL )
	{
		::DestroyWindow( g_hWnd );
		g_hWnd = NULL;
		::UnregisterClass( APPNAME, NULL );
	}
	if( g_hMutex != NULL ){ ::CloseHandle( g_hMutex ); g_hMutex = NULL; }
	return EXIT_SUCCESS;
}

/*******************************************************************************
 *	@fn				BOOL ProcessEvent( void )
 *	@retval			BOOL			TRUE=OK/FALSE=NG
 *	@brief			Cxg̎s
 ******************************************************************************/
BOOL ProcessEvent( void )
{
	MSG	msg;
	// botLPOOŎgpBɂ̃AvŗDŎgp
	while( ::PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) )
	{
		if( !::GetMessage( &msg, NULL, 0, 0 ) )
			return FALSE;
		::TranslateMessage( &msg );
		::DispatchMessage( &msg );
	}
	return TRUE;
}

/*******************************************************************************
 *	@fn				BOOL Init( void )
 *	@retval			BOOL			TRUE=OK/FALSE=NG
 *	@brief			
 ******************************************************************************/
BOOL Init( void )
{
	// gpNX̐ݒƎ擾
	g_pGraph = reinterpret_cast<DL::CDLGraph*>( DL::SetupGraph( APPWIDTH, APPHEIGHT, TRUE ) );
	g_pSound = reinterpret_cast<DL::CDLSound*>( DL::SetupSound( TRUE ) );
	g_pInput = reinterpret_cast<DL::CDLInput*>( DL::SetupInput() );
	g_pDebug = reinterpret_cast<DL::CDLDebug*>( DL::SetupDebug() );

	// 
	DL::Init( g_hWnd, FRAMERATE );

	{
		g_nFontNum = g_pGraph->LoadFont( 2, "lr SVbN", 16 );

		RECT rc;
		rc.left = rc.top = 4; rc.right = 128; rc.bottom = 20*3;
		int nColor = 0xFFFFFF00;
		LPSTR lpStr = " Esc I\n  ړ\n SPACE ]";
		g_pGraph->AddText( 255, 11, g_nFontNum, DL::enumDraw_On, rc, nColor, lpStr );
	}

	g_nGameOver = 0;
	g_nGameCnt = 0;

	g_nBlockX = START_BLOCKX;
	g_nBlockY = START_BLOCKY;

	// Q[ʂ̏
	for( int i = 0; i < TETRIS_HEIGHT; i++ )
		for( int j = 0; j < TETRIS_WIDTH; j++ )
		{
			int n = 0;
			if( j == 0 )					n = 9;
			if( j == TETRIS_WIDTH - 1 )		n = 9;
			if( i == TETRIS_HEIGHT - 1 )	n = 9;

			g_nStage[ i ][ j ] = n;
		}

	// \̊ʒuݒ
	g_nBaseX = ( ( APPWIDTH - ( TETRIS_WIDTH * TETRIS_SIZE ) ) / 2 );
	g_nBaseY = ( ( APPHEIGHT - ( TETRIS_HEIGHT * TETRIS_SIZE ) ) / 2 );

	// `ppl
	for( int i = 0; i < TETRIS_HEIGHT; i++ )
		for( int j = 0; j < TETRIS_WIDTH; j++ )
		{
			DL::D3DPLANE SPlane[ 4 ] =
			{
				{ static_cast<float>( g_nBaseX + ( j + 0 ) * TETRIS_SIZE ), static_cast<float>( g_nBaseY + ( i + 0 ) * TETRIS_SIZE ), 0, 0xff000000 },
				{ static_cast<float>( g_nBaseX + ( j + 1 ) * TETRIS_SIZE ), static_cast<float>( g_nBaseY + ( i + 0 ) * TETRIS_SIZE ), 0, 0xff000000 },
				{ static_cast<float>( g_nBaseX + ( j + 1 ) * TETRIS_SIZE ), static_cast<float>( g_nBaseY + ( i + 1 ) * TETRIS_SIZE ), 0, 0xff000000 },
				{ static_cast<float>( g_nBaseX + ( j + 0 ) * TETRIS_SIZE ), static_cast<float>( g_nBaseY + ( i + 1 ) * TETRIS_SIZE ), 0, 0xff000000 },
			};
			g_pGraph->AddPlane( i * TETRIS_WIDTH + j, 10, -1, DL::enumDraw_On, DL::enumState_Normal, 4, SPlane );
		}

	// ubN쐬
	CreateBlock();

	return TRUE;
}

/*******************************************************************************
 *	@fn				void KeyState( BYTE* byKey )
 *	@param[in]		byKey						L[obt@
 *	@retval			none						Ȃ
 *	@brief			L[
 ******************************************************************************/
void KeyState( BYTE* byKey )
{
	::memcpy_s( g_byKey, DL::CDLInput::KEYBUFFER_MAX, byKey, DL::CDLInput::KEYBUFFER_MAX );

	if( g_byKey[ DIK_ESCAPE ] & 0x80 )											//!< ESCL[
	{
		::DestroyWindow( g_hWnd );
	}
	// Q[̂݃L[󂯕t
	if( g_nGameOver == 0 )
	{
		if( g_byKey[ DIK_LEFT ] == 0 ) g_bKey[ 0 ] = TRUE;
		if( g_byKey[ DIK_RIGHT ] == 0 )g_bKey[ 1 ] = TRUE;
		if( g_byKey[ DIK_SPACE ] == 0 )g_bKey[ 3 ] = TRUE;

		//!< ړ
		if( g_byKey[ DIK_LEFT ] & 0x80 )
		{
			if( g_bKey[ 0 ] )
			{
				if( CheckOverlap( g_nBlockX - 1, g_nBlockY ) )
					MoveBlock( g_nBlockX - 1, g_nBlockY );
				g_bKey[ 0 ] = FALSE;
			}
		}else
		//!< Eړ
		if( g_byKey[ DIK_RIGHT ] & 0x80 )
		{
			if( g_bKey[ 1 ] )
			{
				if( CheckOverlap( g_nBlockX + 1, g_nBlockY ) )
					MoveBlock( g_nBlockX + 1, g_nBlockY );
				g_bKey[ 1 ] = FALSE;
			}
		}else
		//!< ړ
		if( g_byKey[ DIK_DOWN ] & 0x80 )
		{
			if( CheckOverlap( g_nBlockX, g_nBlockY + 1 ) )
				MoveBlock( g_nBlockX, g_nBlockY + 1 );
		}else
		// !< ]
		if( g_byKey[ DIK_SPACE ] & 0x80 )
		{
			if( g_bKey[ 3 ] )
			{
				TurnBlock();
				g_bKey[ 3 ] = FALSE;
			}
		}
	}
}

/*******************************************************************************
 *	@fn				void Proc( BYTE* byKey,
 *							   DIMOUSESTATE* diMouse )
 *	@param[in]		byKey						L[obt@
 *	@param[in]		diMouse						}EXobt@
 *	@retval			none						Ȃ
 *	@brief			vO
 ******************************************************************************/
void Proc( BYTE* byKey, DIMOUSESTATE* diMouse )
{
	// L[
	KeyState( byKey );

	if( g_nGameOver == 0 )
	{
		if( g_nGameCnt++ > 30 )
		{
			DropBlock();
			g_nGameCnt = 0;
		}
	}

	// `
	DrawTetris();
}

/*******************************************************************************
 *	@fn				int WINAPI WinMain( HINSTANCE hInstance,
 *									    HINSTANCE hPrevInstance,
 *										LPSTR lpCmdLine,
 *										int nCmdShow )
 *	@param[in]		hInstance		CX^Xnh
 *	@param[in]		hPrevInstance	CX^Xnh
 *	@param[in]		lpCmdLine		
 *	@param[in]		nCmdShow		
 *	@retval			LRESULT			
 *	@brief			vON
 ******************************************************************************/
int WINAPI WinMain( HINSTANCE hInstance,
				    HINSTANCE hPrevInstance,
					LPSTR lpCmdLine,
					int nCmdShow )
{
	// AvP[V̏
	if( Init( hInstance ) == EXIT_FAILURE ) return EXIT_FAILURE;
	// DirectXLibrary̏
	if( !Init() ) goto ERR_EXIT;

	// [v
	do
	{
		DL::Exec( Proc );
	} while( ProcessEvent() );

ERR_EXIT:
	// DirectXLibrary̏I
	DL::End();
	// AvP[V̏I
	Exit();

	return EXIT_SUCCESS;
}
